/**
 * @file backend_d.d
 * @brief Representation of a generated D program
 * @author Sebastien Alaiwan - Romain Bouqueau
 * @date 2014-12-16
 */

import codestream;
import std.string;
import backend;

class Backend_D : Backend
{
  void printCode(const GProgram p, CodeStream s)
  {
    foreach(gstruct; p.structs)
      printStruct(gstruct, s);

    foreach(gfunction; p.functions)
      printFunction(gfunction, s);
  }

private:
  void printStruct(const GStruct gstruct, CodeStream s)
  {
    assert(gstruct.name != "");
    s.writefln("struct %s", gstruct.name);
    s.writefln("{");
    s.indent();

    foreach(memberDesc; gstruct.members)
      printMember(memberDesc, s);

    s.unindent();
    s.writefln("}");
    s.writefln("");
  }

  void printFunction(const GFunction gfunction, CodeStream s)
  {
    const retType = firstNotEmpty(gfunction.retType, "void");

    s.writefln("%s %s(%s)", retType, gfunction.name,
               myJoin(gfunction.args, ", "));
    s.writefln("{");
    s.indent();

    if(gfunction.retType != "")
      s.writefln("%s r;", gfunction.retType);

    foreach(line; gfunction.codelines)
      printCodeline(line, s);

    if(gfunction.retType != "")
      s.writefln("return r;");

    s.unindent();
    s.writefln("}");
    s.writefln("");
  }

  void printCodeline(const GCodeline codeline, CodeStream s)
  {
    void printCallFunction(const(GCodelineCallFunction)cf, CodeStream s)
    {
      s.writefln(cf.codeline);
    }

    void printOperation(const(GCodelineOperation)co, CodeStream s)
    {
      s.writefln("auto " ~ co.left ~ " " ~ co.operator ~ " " ~ co.right ~ ";");
    }

    codeline.visit(&printCallFunction, &printOperation,
                   s);
  }

  string firstNotEmpty(string a, string b)
  {
    if(a != "")
      return a;
    else
      return b;
  }

  string myJoin(const string[] tab, string sep)
  {
    string r;

    foreach(i, element; tab)
    {
      if(i > 0)
        r ~= sep;

      r ~= element;
    }

    return r;
  }

  void printMember(const GStructMember desc, CodeStream s)
  {
    string line;
    line ~= format("%s %s", desc.type, desc.name);

    if(desc.arrayLength > 1)
      line ~= format("[%s]", desc.arrayLength);

    line ~= ";";

    if(desc.comment != "")
      line ~= format("// %s", desc.comment);

    s.writefln("%s", line);
  }
}

