import model;
import codestream;
import backend;
import std.stdio;
import std.string;

string getParserCode(MODEL m, Backend b)
{
  const program = generateParser(m);

  auto s = new CodeStream;
  b.printCode(program, s);
  return s.toString();
}

GProgram generateParser(MODEL m)
{
  auto prog = new GProgram;

  Context ctx;

  foreach(block; m)
    generateBlock(prog, block, ctx);

  return prog;
}

private:
struct Context
{
  GStruct currStruct;
  GFunction currParseFunc;
  GFunction currPrintFunc;
}

void generateBlock(GProgram prog, in Block block, ref Context ctx)
{
  void doSequence(const(SequenceBlock)desc)
  {
    foreach(stmt; desc.statements)
      generateBlock(prog, stmt, ctx);
  }

  void doNamedSequence(const(NamedSequenceBlock)desc)
  {
    auto gstruct = new GStruct;
    gstruct.name = desc.sName;
    prog.structs ~= gstruct;
    ctx.currStruct = gstruct;

    if(ctx.currParseFunc)
    {
      ctx.currParseFunc.codelines ~=
      mkCallFunction(format("parse_%s(stream);", desc.sName));
    }

    {
      auto parseFunc = new GFunction;
      prog.functions ~= parseFunc;
      ctx.currParseFunc = parseFunc;

      parseFunc.retType = desc.sName;
      parseFunc.name = "parse_" ~ desc.sName;
      parseFunc.args = ["Stream stream", ];
    }

    {
      auto printFunc = new GFunction;
      prog.functions ~= printFunc;
      ctx.currPrintFunc = printFunc;

      printFunc.name = "print_" ~ desc.sName;
      printFunc.args = ["string name", desc.sName ~ " s", "File o"];
      printFunc.codelines ~=
      mkCallFunction(format(`writeln("- %s");`, desc.sName));
    }

    doSequence(desc.seq);
  }

  void doField(const FieldBlock desc)
  {
    string formatParse(const FieldBlock desc)
    {
      string line;

      if(desc.iSizeInBits < 0)
      {
        line ~= format("parse_%s(stream);", desc.sType);
      }
      else
      {
        line ~= format("parse_%s(stream, %s);",
                       desc.sType, desc.iSizeInBits);
      }

      return line;
    }

    bool isInArray(const ref GStructMember m, const ref GStructMember[] members)
    {
      foreach(ref member; members)
      {
        if(m.name == member.name)
        {
          return true;
        }
      }

      return false;
    }

    auto member = new GStructMember;
    member.type = desc.sType;
    member.name = desc.sName;
    member.arrayLength = desc.iArrayLength;
    member.comment = desc.sComment;

    if(isInArray(member, ctx.currStruct.members))
      writeln("Found a duplicate entry for \"", member.name, "\".");
    else
      ctx.currStruct.members ~= member;

    if(desc.iArrayLength > 1)
    {
      ctx.currParseFunc.codelines ~=
      mkCallFunction(format("for (size_t i=0; i<%s; ++i)",
                            desc.iArrayLength));
      ctx.currParseFunc.codelines ~=
      mkCallFunction(format("  r.%s[i] = %s", desc.sName, formatParse(
                              desc)));
    }
    else
    {
      ctx.currParseFunc.codelines ~=
      mkCallFunction(format("r.%s = %s", desc.sName, formatParse(
                              desc)));
    }

    if(desc.iArrayLength > 1)
    {
      ctx.currPrintFunc.codelines ~=
      mkCallFunction(format("for(size_t i=0; i<%s; ++i)",
                            desc.iArrayLength));
      ctx.currPrintFunc.codelines ~=
      mkCallFunction(format(`  print_%s("%s", s.%s[i], o);`,
                            desc.sType,
                            desc.sName, desc.sName));
    }
    else
    {
      ctx.currPrintFunc.codelines ~=
      mkCallFunction(format(`print_%s("%s", s.%s, o);`,
                            desc.sType,
                            desc.sName, desc.sName));
    }
  }

  void doOperation(const OperationBlock operationBlock)
  {
    auto parseFunc = ctx.currParseFunc;

    parseFunc.codelines ~= mkOperation(operationBlock.sLeftOperation,
                                       operationBlock.sOperator,
                                       operationBlock.sRightOperation);
  }

  void doIf(const IfBlock ifBlock)
  {
    auto parseFunc = ctx.currParseFunc;

    parseFunc.codelines ~= mkCallFunction("if(" ~ ifBlock.condition ~ ")");
    parseFunc.codelines ~= mkCallFunction("{");
    ctx.currParseFunc = parseFunc;
    generateBlock(prog, ifBlock.m_then, ctx);
    parseFunc.codelines ~= mkCallFunction("}");
    parseFunc.codelines ~= mkCallFunction("else");
    parseFunc.codelines ~= mkCallFunction("{");
    ctx.currParseFunc = parseFunc;
    generateBlock(prog, ifBlock.m_else, ctx);
    parseFunc.codelines ~= mkCallFunction("}");
  }

  void doLoop(const LoopBlock loopBlock)
  {
    auto parseFunc = ctx.currParseFunc;

    parseFunc.codelines ~= mkCallFunction("while(" ~ loopBlock.condition ~ ")");
    parseFunc.codelines ~= mkCallFunction("{");
    ctx.currParseFunc = parseFunc;
    generateBlock(prog, loopBlock.m_body, ctx);
    parseFunc.codelines ~= mkCallFunction("}");
  }

  block.visit(
    &doNamedSequence,
    &doSequence,
    &doIf,
    &doLoop,
    &doField,
    &doOperation);
}

